#
#  This file is part of bioservices software
#
#  Copyright (c) 2013-2017 - EBI-EMBL
#
#  File author(s): Nick Weiner and others
#
#  Distributed under the GPLv3 License.
#  See accompanying file LICENSE.txt or copy at
#      http://www.gnu.org/licenses/gpl-3.0.html
#
#  website: https://github.com/cokelaer/bioservices
#  documentation: http://packages.python.org/bioservices
#
##############################################################################
"""This module provides a class :class:`~BioModels` that allows an easy access
to all the BioModel service.


.. topic:: What is BioMart ?

    :URL: http://www.biomart.org/
    :REST: http://www.biomart.org/martservice.html

    .. highlights::

        The BioMart project provides free software and data services to the
        international scientific community in order to foster scientific collaboration
        and facilitate the scientific discovery process. The project adheres to the open
        source philosophy that promotes collaboration and code reuse.

        -- from BioMart March 2013

.. note:: SOAP and REST are available. We use REST for the wrapping.
"""
from io import StringIO
from bioservices import REST, BioServicesError
from functools import wraps
from bioservices import logger

logger.name = __name__
import pandas as pd

__all__ = ["BioMart"]


def require_host(f):
    @wraps(f)
    def wrapper(*args, **kargs):
        if args[0].host is None:
            print("You must set the host (e.g. f.host='www.ensembl.org' ")
            return
        return f(*args, **kargs)

    return wrapper


class BioMart(REST):
    r"""Interface to the `BioMart <http://www.biomart.org>`_ service

    BioMart is made of different views. Each view correspond to a specific **MART**.
    For instance the UniProt service has a `BioMart view <http://www.ebi.ac.uk/uniprot/biomart/martview/>`_.

    The registry can help to find the different services available through
    BioMart.

        >>> from bioservices import *
        >>> s = BioMart()
        >>> ret = s.registry() # to get information about existing services

    The registry is a list of dictionaries. Some aliases are available to get
    all the names or databases::

        >>> s.names      # alias to list of valid service names from registry
        >>> "unimart" in s.names
        True

    Once you selected a view, you will want to select a database associated with
    this view and then a dataset. The datasets can be retrieved as follows::

        >>> s.datasets("prod-intermart_1")  # retrieve datasets available for this mart

    The main issue is how to figure out the database name (here **prod-intermart_1**) ?
    Indeed, from the web site, what you see is the **displayName** and you must
    introspect the registry to get this information. In **BioServices**, we provide
    the :meth:`~bioservices.biomart.BioMart.lookfor` method to help you. For instance, to
    retrieve the database name of **interpro**, type::

        >>> s = BioMart(verbose=False)
        >>> s.lookfor("interpro")
        Candidate:
             database: intermart_1
            MART name: prod-intermart_1
          displayName: INTERPRO (EBI UK)
                hosts: www.ebi.ac.uk

    The display name (INTERPRO) correspond to the MART name
    prod-intermart_1. Let us you it to retrieve the datasets::

        >>> s.datasets("prod-intermart_1")
        ['protein', 'entry', 'uniparc']

    Now that we have the dataset names, we can select one and build a
    query. Queries are XML that contains the dataset name, some
    attributes and filters. The dataset name is one of the element
    returned by the datasets method. Let us suppose that we want to query
    **protein**, we need to add this dataset to the query::

        >>> s.add_dataset_to_xml("protein")

    Then, you can add attributes (one of the keys of the dictionary
    returned by attributes("protein")::

        >>> s.add_attribute_to_xml("protein_accession")

    Optional filters can be used::

        >>> s.add_filter_to_xml("protein_length_greater_than", 1000)

    Finally, you can retrieve the XML query::

        >>> xml_query = s.get_xml()

    and send the request to biomart::

        >>> res = s.query(xml_query)
        >>> len(res)
        12801
        # print the first 10 accession numbers
        >>> res = res.split("\n")
        >>> for x in res[0:10]: print(x)
        ['P18656',
         'Q81998',
         'O09585',
         'O77624',
         'Q9R3A1',
         'E7QZH5',
         'O46454',
         'Q9T3F4',
         'Q9TCA3',
         'P72759']


    REACTOME example::

        s.lookfor("reactome")
        s.datasets("REACTOME")
        ['interaction', 'complex', 'reaction', 'pathway']

        s.new_query()
        s.add_dataset_to_xml("pathway")
        s.add_filter_to_xml("species_selection", "Homo sapiens")
        s.add_attribute_to_xml("pathway_db_id")
        s.add_attribute_to_xml("_displayname")
        xmlq = s.biomartQuery.get_xml()
        res = s.query(xmlq)

    .. note:: the biomart sevice is slow (in my experience, 2013-2014) so please be patient...

    """
    _xml_example = """<?xml version="1.0" encoding="UTF-8"?>
                <!DOCTYPE Query>
                    <Query virtualSchemaName="default" formatter="CSV" header="0" uniqueRows="0" count="" datasetConfigVersion="0.6">
                    <Dataset name="mmusculus_gene_ensembl" interface="default">
                    <Filter name="ensembl_gene_id" value="ENSMUSG00000086981"/>
                    <Attribute name="ensembl_gene_id"/>
                    <Attribute name="ensembl_transcript_id"/>
                    <Attribute name="transcript_start"/>
                    <Attribute name="transcript_end"/>
                    <Attribute name="exon_chrom_start"/>
                    <Attribute name="exon_chrom_end"/>
                    </Dataset>
                    </Query>"""

    def __init__(self, host=None, verbose=False, cache=False, secure=False):
        """.. rubric:: Constructor


        URL required to use biomart change quite often. Experience has
        shown that BioMart class in Bioservices may fail. This is not a
        bioservices issue but due to API changes on server side.

        For that reason the host is not filled anymore and one must set it
        manually.

        Let us take the example of the ensembl biomart. The host is

            www.ensembl.org

        Note that there is no prefix *http* and that the actual URL looked for
        internally is http://www.ensembl.org/biomart/martview

        (It used to be martservice in 2012-2016)

        Another reason to not set any default host is that servers may be busy or
        take lots of time to initialise (if many MARTS are available). Usually,
        one knows which MART to look at, in which case you may want to use a
        specific host (e.g., www.ensembl.org) that will speed up significantly the
        initialisation time.

        :param str host: a valid host (e.g. "www.ensembl.org", gramene.org)

        List of databases are available in this webpage http://www.biomart.org/community.html


        """
        url = "undefined"
        super(BioMart, self).__init__("BioMart", url=url, verbose=verbose, cache=cache, url_defined_later=True)

        self._names = None
        self._marts = None
        self._databases = None
        self._display_names = None
        self._valid_attributes = None
        self._hosts = None
        self._host = None
        self._secure = secure

        if host is None:
            i = 0
            hosts = ["www.ensembl.org", "asia.ensembl.org", "useast.ensembl.org"]
            while self.host is None and (i < 3):
                self.host = hosts[i]
                i += 1
            if self.host is None:
                raise IOError("no host provided and no default hosts {} not reachable".format(hosts))
        else:
            self.host = host
        self._biomartQuery = BioMartQuery()

    def custom_query(self, **args):
        self._biomartQuery = BioMartQuery(**args)

    def _get_host(self):
        return self._host

    def _set_host(self, host):
        import requests

        secure = ""
        if self._secure:
            secure = "s"
        url = "http{}://{}/biomart/martservice".format(secure, host)
        request = requests.head(url)
        if request.status_code in [200]:
            self._host = host
            self.url = url
            self._init()
        else:
            self.logging.warning("host {} is not reachable ".format(host))

    host = property(_get_host, _set_host)

    def _set_format(self, format):
        self.format = format

    def _init(self):
        temp = self.logging.level
        self.logging.setLevel("ERROR")
        _ = self.lookfor("uniprot", verbose=False)
        _ = self.valid_attributes
        self.logging.setLevel(temp)

    @require_host
    def registry(self):
        """to retrieve registry information

         the XML contains list of children called MartURLLocation made
         of attributes. We parse the xml to return a list of dictionary.
         each dictionary correspond to one MART.

        aliases to some keys are provided: names, databases, displayNames

        """
        ret = self.http_get("?type=registry", frmt="xml")
        ret = self.easyXML(ret)
        # the XML contains list of children called MartURLLocation made
        # of attributes. We parse the xml to return a list of dictionary.
        # each dictionary correspond to one MART.
        ret = [x.attrib for x in ret.getchildren()]
        return ret

    @require_host
    def datasets(self, mart, raw=False):
        """to retrieve datasets available for a mart:

        :param str mart: e.g. ensembl. see :attr:`names` for a list of valid
            MART names the mart is the database. see lookfor method or
            databases attributes

        >>> s = BioMart(verbose=False)
        >>> s.datasets("prod-intermart_1")
        ['protein', 'entry', 'uniparc']

        """
        if mart not in self.names:
            raise BioServicesError("Provided mart name (%s) is not valid. see 'names' attribute" % mart)
        ret = self.http_get("?type=datasets&mart=%s" % mart, frmt="txt")

        if raw is False:
            try:
                ret2 = [x.split("\t") for x in ret.split("\n") if len(x.strip())]
                ret = [x[1] for x in ret2]
            except:
                ret = ["?"]
        return ret

    def get_datasets(self, mart):
        """Retrieve datasets with description"""
        if mart not in self.names:
            raise BioServicesError("Provided mart name (%s) is not valid. see 'names' attribute" % mart)

        ret = self.http_get("?type=datasets&mart=%s" % mart, frmt="txt")
        import pandas as pd

        df = pd.read_csv(
            StringIO(ret),
            sep="\t",
            header=None,
            usecols=[1, 2],
            names=["name", "description"],
        )
        return df

    @require_host
    def attributes(self, dataset):
        """to retrieve attributes available for a dataset:

        :param str dataset: e.g. oanatinus_gene_ensembl

        """
        # assert dataset in self.names
        if dataset not in [x for k in self.valid_attributes.keys() for x in self.valid_attributes[k]]:
            raise ValueError("provided dataset (%s) is not found. see valid_attributes" % dataset)
        ret = self.http_get("?type=attributes&dataset=%s" % dataset, frmt="txt")

        ret = [x for x in ret.split("\n") if len(x)]
        results = {}
        for line in ret:
            key = line.split("\t")[0]
            results[key] = line.split("\t")[1:]
        return results

    @require_host
    def filters(self, dataset):
        r"""to retrieve filters available for a dataset:

        :param str dataset: e.g. oanatinus_gene_ensembl

        ::

            >>> s.filters("uniprot").split("\n")[1].split("\t")
            >>> s.filters("pathway")["species_selection"]
            [Arabidopsis thaliana,Bos taurus,Caenorhabditis elegans,Canis familiaris,Danio
            rerio,Dictyostelium discoideum,Drosophila melanogaster,Escherichia coli,Gallus
            gallus,Homo sapiens,Mus musculus,Mycobacterium tuberculosis,Oryza
            sativa,Plasmodium falciparum,Rattus norvegicus,Saccharomyces
            cerevisiae,Schizosaccharomyces pombe,Staphylococcus aureus N315,Sus
            scrofa,Taeniopygia guttata ,Xenopus tropicalis]

        """
        if dataset not in [x for k in self.valid_attributes.keys() for x in self.valid_attributes[k]]:
            raise ValueError("provided dataset (%s) is not found. see valid_attributes" % dataset)
        ret = self.http_get("?type=filters&dataset=%s" % dataset, frmt="txt")
        ret = [x for x in ret.split("\n") if len(x)]
        results = {}
        for line in ret:
            key = line.split("\t")[0]
            results[key] = line.split("\t")[1:]
        return results

    @require_host
    def configuration(self, dataset):
        """to retrieve configuration available for a dataset:

        :param str dataset: e.g. oanatinus_gene_ensembl

        """
        ret = self.http_get("?type=configuration&dataset=%s" % dataset, frmt="xml")
        ret = self.easyXML(ret)
        return ret

    @require_host
    def version(self, mart):
        """Returns version of a **mart**

        :param str mart: e.g. ensembl

        """
        ret = self.http_get("?type=version&mart=%s" % mart, frmt="xml")
        ret = self.easyXML(ret)
        return ret.root.strip()

    @require_host
    def new_query(self):
        self._biomartQuery.reset()

    @require_host
    def query(self, xmlq):
        """Send a query to biomart

        The query must be formatted in a XML format which looks like (
        example from https://gist.github.com/keithshep/7776579)::

            <?xml version="1.0" encoding="UTF-8"?>
                <!DOCTYPE Query>
                    <Query virtualSchemaName="default" formatter="CSV" header="0" uniqueRows="0" count="" datasetConfigVersion="0.6">
                    <Dataset name="mmusculus_gene_ensembl" interface="default">
                    <Filter name="ensembl_gene_id" value="ENSMUSG00000086981"/>
                    <Attribute name="ensembl_gene_id"/>
                    <Attribute name="ensembl_transcript_id"/>
                    <Attribute name="transcript_start"/>
                    <Attribute name="transcript_end"/>
                    <Attribute name="exon_chrom_start"/>
                    <Attribute name="exon_chrom_end"/>
                    </Dataset>
                    </Query>

        .. warning:: the input XML must be valid. THere is no validation made
            in thiss method.
        """
        ret = self.http_post(None, frmt=None, data={"query": xmlq.strip()}, headers={})
        return ret

    @require_host
    def add_attribute_to_xml(self, name, dataset=None):
        attr = self.create_attribute(name, dataset)
        self._biomartQuery.add_attribute(attr)

    @require_host
    def add_filter_to_xml(self, name, value, dataset=None):
        filt = self.create_filter(name, value, dataset)
        self._biomartQuery.add_filter(filt)

    @require_host
    def add_dataset_to_xml(self, dataset):
        self.attributes(dataset)
        #    raise BioServicesError("invalid dataset names provided. Check names attribute")
        self._biomartQuery.add_dataset(dataset)

    @require_host
    def get_xml(self):
        return self._biomartQuery.get_xml()

    @require_host
    def create_filter(self, name, value, dataset=None):
        if dataset:
            valid_filters = self.filters(dataset).keys()
            if name not in valid_filters:
                raise BioServicesError("Invalid filter name. ")
        _filter = ""
        if "=" in value:
            _filter = """        <Filter name = "%s" %s/>""" % (name, value)
        else:
            _filter = """        <Filter name = "%s" value = "%s"/>""" % (name, value)
        return _filter

    @require_host
    def create_attribute(self, name, dataset=None):
        # s.attributes(dataset)
        # valid dataset
        if dataset:
            valid_attributes = self.attributes(dataset).keys()
            if name not in valid_attributes:
                raise BioServicesError("Invalid attribute name. ")
        attrib = """        <Attribute name = "%s" />""" % name
        return attrib

    @require_host
    def _get_names(self):
        if self._names is None:
            ret = self.registry()
            names = [x["name"] for x in ret]
            self._names = names[:]
        return self._names

    names = property(_get_names, doc="list of valid datasets")

    @require_host
    def _get_displayNames(self):
        if self._display_names is None:
            ret = self.registry()
            names = [x["displayName"] for x in ret]
            self._display_names = names[:]
        return self._display_names

    displayNames = property(_get_displayNames, doc="list of valid datasets")

    @require_host
    def _get_databases(self):
        if self._databases is None:
            ret = self.registry()
            names = [x.get("database", "?") for x in ret]
            self._databases = names
        return self._databases

    databases = property(_get_databases, doc="list of valid datasets")

    @require_host
    def _get_marts(self):
        if self._marts is None:
            ret = self.registry()
            df = pd.DataFrame(ret)[["database", "displayName", "name"]]
            self._marts = df
        return self._marts

    marts = property(_get_marts, doc="list of marts")

    @require_host
    def _get_hosts(self):
        if self._hosts is None:
            ret = self.registry()
            names = [x.get("host", "?") for x in ret]
            self._hosts = names[:]
        return self._hosts

    hosts = property(_get_hosts, doc="list of valid hosts")

    @require_host
    def _get_valid_attributes(
        self,
    ):
        res = {}
        if self._valid_attributes is None:
            # we could use a loop and call self.datasets(name, raw=False) but it
            # can be a bit longish, so we use the asynchronous call using
            # requests
            saveme = self.settings.params["general.async_threshold"]
            # TODO: not python3 compatible for now. Waiting for gevent package
            # to be available.
            self.settings.params["general.async_threshold"][0] = 10000  #

            queries = ["?type=datasets&mart=%s" % name for name in self.names]
            results = self.http_get(queries, frmt="txt")

            self.settings.params["general.async_threshold"] = saveme
            # requests.start()
            # requests.wait()
            # results = requests.get_results()

            for i, name in enumerate(self.names):
                try:
                    res[name] = [x.split("\t")[1] for x in results[i].split("\n") if len(x.strip()) > 1]
                except:
                    res[name] = "?"
            self._valid_attributes = res.copy()
        return self._valid_attributes

    valid_attributes = property(_get_valid_attributes, doc="list of valid datasets")

    @require_host
    def lookfor(self, pattern, verbose=True):
        for a, x, y, z in zip(self.hosts, self.databases, self.names, self.displayNames):
            found = False
            if pattern.lower() in x.lower():
                found = True
            if pattern.lower() in y.lower():
                found = True
            if pattern.lower() in z.lower():
                found = True
            if found is True and verbose is True:
                print("Candidate:")
                print("     database: %s " % x)
                print("    MART name: %s " % y)
                print("  displayName: %s " % z)
                print("        hosts: %s " % a)


class BioMartQuery(object):
    def __init__(
        self,
        version="1.0",
        virtualScheme="default",
        formatter="TSV",
        header=0,
        unique=0,
        configVer="0.6",
    ):
        params = {
            "version": version,
            "virtualSchemaName": virtualScheme,
            "formatter": formatter,
            "header": header,
            "uniqueRows": unique,
            "configVersion": configVer,
        }

        self.header = (
            """<?xml version="%(version)s" encoding="UTF-8"?>
<!DOCTYPE Query>
<Query  virtualSchemaName = "%(virtualSchemaName)s" formatter = "%(formatter)s"
header = "%(header)s" uniqueRows = "%(uniqueRows)s" count = ""
datasetConfigVersion = "%(configVersion)s" >\n"""
            % params
        )

        self.footer = "    </Dataset>\n</Query>"
        self.reset()

    def add_filter(self, filter):
        self.filters.append(filter)

    def add_attribute(self, attribute):
        self.attributes.append(attribute)

    def add_dataset(self, dataset):
        self.dataset = """    <Dataset name = "%s" interface = "default" >""" % dataset

    def reset(self):
        self.attributes = []
        self.filters = []
        self.dataset = None

    def get_xml(self):
        if self.dataset is None:
            raise BioServicesError("data set must be set. Use add_dataset method")
        xml = self.header
        xml += self.dataset + "\n\n"
        for line in self.filters:
            xml += line + "\n"
        for line in self.attributes:
            xml += line + "\n"
        xml += self.footer
        return xml



# imports - standard imports
import sys

PYTHON_VERSION = sys.version_info


def _is_python_version(*args, **kwargs):
    major = kwargs.get("major", None)
    minor = kwargs.get("minor", None)
    patch = kwargs.get("patch", None)

    result = True

    if major:
        result = result and major == PYTHON_VERSION.major
    if minor:
        result = result and minor == PYTHON_VERSION.minor
    if patch:
        result = result and patch == PYTHON_VERSION.micro

    return result


PY2 = _is_python_version(major=2)


def iteritems(dict_, **kwargs):
    if PY2:
        iterator = dict_.iteritems()
    else:
        iterator = iter(dict_.items(), **kwargs)
    return iterator


if PY2:
    string_types = basestring
else:
    string_types = str



#
#  This file is part of bioservices software
#
#  Copyright (c) 2013-2014 - EBI-EMBL
#
#
#  Distributed under the GPLv3 License.
#  See accompanying file LICENSE.txt or copy at
#      http://www.gnu.org/licenses/gpl-3.0.html
#
#  website: https://github.com/cokelaer/bioservices
#  documentation: http://packages.python.org/bioservices
#
##############################################################################
"""This module provides a class :class:`ENA`

.. topic:: What is ENA

    :URL:  https://www.ebi.ac.uk/ena


    .. highlights::

        The European Nucleotide Archive (ENA) provides a comprehensive
        record of the world's nucleotide sequencing information, covering
        raw sequencing data, sequence assembly information and functional
        annotation.

        -- From ENA web page Jan 2016

.. versionadded:: 1.4.4

"""
import os
from bioservices.services import REST
import webbrowser
from bioservices import logger

logger.name = __name__


__all__ = ["ENA"]


class ENA:
    """Interface to `ChEMBL <http://www.ebi.ac.uk/ena/index.php>`_

    Here is a quick example to retrieve a target given its ChEMBL Id

    .. doctest::

        >>> from bioservices import ENA
        >>> s = ENA(verbose=False)


    Retrieve read domain metadata in XML format::

        print(e.get_data('ERA000092', 'xml'))

    Retrieve assemble and annotated sequences in fasta format::

        print(e.get_data('A00145', 'fasta'))

    The range parameter can be used in combination to retrieve a subsequence
    from sequence entry A00145 from bases 3 to 63 using ::

        e.get_data('A00145', 'fasta', fasta_range=[3,63])

    Retrieve assembled and annotated subsequences in HTML format (same
    as above but in HTML page).

        e.view_data('A00145')


    Retrieve expanded CON records:

    To retrieve expanded CON records use the expanded=true parameter. For
    example, the expanded CON entry AL513382 in flat file format can be i
    obtained as follows::

        e.get_data('AL513382', frmt='text', expanded=True)

    Expanded CON records are different from CON records in two ways.
    Firstly, the expanded CON records contain the full sequence in addition
    to the contig assembly instructions. Secondly, if a CON record contains
    only source or gap features the expanded CON records will also display
    all features from the segment records.

    Retrieve assembled and annotated sequence header in flat file format

    To retrieve assembled and annotated sequence header in flat file
    format please use the header=true parameter, e.g.:

        e.get_data('BN000065', 'text', header=True)


    Retrieve assembled and annotated sequence records using sequence
    versions::

        e.get_data('AM407889.1', 'fasta')
        e.get_data('AM407889.2', 'fasta')


    """

    url = "http://www.ebi.ac.uk/ena/browser/api"

    def __init__(self, verbose=False, cache=False):
        """**Constructor**

        :param verbose: set to False to prevent informative messages
        """
        self.services = REST(name="ENA", url=ENA.url, verbose=verbose, cache=cache)
        self.services.TIMEOUT = 100

    def get_data(
        self,
        identifier,
        frmt,
        fasta_range=None,
        expanded=None,
        header=None,
        download=None,
    ):
        """

        :param frmt : xml, text, fasta, fastq, html, embl but does depend on the
            entry

        Example:

            get_data("/AL513382", "embl")

        ENA API changed in 2020 but we tried to keep the same services in this
        method.
        """

        url = f"{self.url}/{frmt}/{identifier}"

        if frmt in ["text", "fasta", "fastq"]:
            res = self.services.http_get(url, frmt="txt")
        elif frmt in ["html"]:
            res = self.services.http_get(url, frmt="default")
        elif frmt in ["xml"]:
            res = self.services.http_get(url, frmt="xml")
        return res

    def data_warehouse(self):
        # http://www.ebi.ac.uk/ena/data/warehouse/search?query="geo_circ(-0.587,-90.5713,170)"&result=sequence_release&display=text&download=gzip
        pass

    def get_taxon(self, taxon):
        print("deprecated since v.7.8 due to ENA update")



import pkg_resources

from easydev import CustomConfig
from easydev.logging_tools import Logging


logger = Logging("bioservices", "WARNING", text_color="green")

try:
    version = pkg_resources.require("bioservices")[0].version
except:
    version = "1.9.0"

import colorlog

logger = colorlog.getLogger(logger.name)

# Initialise the config directory if not already done
configuration = CustomConfig("bioservices", verbose=False)
bspath = configuration.user_config_dir

# Add bioservices.uniprot to sys.modules to prevent cycles in our imports
# import bioservices.uniprot
# bioservices.uniprot  # Stop flake8 error

from . import settings
from .settings import *

from . import services
from .services import *

from . import biodbnet
from .biodbnet import *

from . import biomodels
from .biomodels import *

from . import biocontainers
from .biocontainers import Biocontainers

from . import cog
from .cog import *

from . import chebi
from .chebi import *

from . import ena
from .ena import ENA

from . import eva
from .eva import *

from . import ensembl
from .ensembl import *

# moved to attic in bioservices v1.6
# from . import geneprof
# from .geneprof import *

from . import kegg
from .kegg import *

from . import hgnc
from .hgnc import *

from . import intact
from .intact import *

from . import pubchem
from .pubchem import *

from . import pfam
from .pfam import *

from . import rhea
from .rhea import *

from . import xmltools
from .xmltools import *

from . import wikipathway
from .wikipathway import *

from . import omnipath
from .omnipath import *

from . import pdb
from .pdb import *

from . import pdbe
from .pdbe import *

from . import pride
from .pride import *

from . import uniprot
from .uniprot import *

from . import unichem
from .unichem import *

from . import rnaseq_ebi
from .rnaseq_ebi import RNASEQ_EBI

from . import reactome
from .reactome import *

from . import quickgo
from .quickgo import *


from . import chembl
from .chembl import *

# from . import picr
# from .picr import *

from . import psicquic
from .psicquic import *

from . import ncbiblast
from .ncbiblast import *

# from . import readseq
# from .readseq import *

from . import biogrid
from .biogrid import *

from . import arrayexpress
from .arrayexpress import *

from . import biomart
from .biomart import *

from . import eutils
from .eutils import *

from . import pathwaycommons
from .pathwaycommons import *

from . import muscle
from .muscle import *

from . import dbfetch
from .dbfetch import *

from . import bigg
from .bigg import BiGG

from . import omicsdi
from .omicsdi import OmicsDI

# sub packages inside bioservices.
from . import apps



import os
from setuptools import setup, find_packages
import glob


_MAJOR               = 1
_MINOR               = 10
_MICRO               = 3
version              = '%d.%d.%d' % (_MAJOR, _MINOR, _MICRO)
release              = '%d.%d' % (_MAJOR, _MINOR)

metainfo = {
    'authors': {
        'Cokelaer':('Thomas Cokelaer','thomas.cokelaer@pasteur.fr'),
        },
    'version': version,
    'license' : 'GPLv3',
    'download_url' : 'http://pypi.python.org/pypi/bioservices',
    'url' : 'http://github.com/cokelaer/bioservices',
    'bugtrack_url': 'https://github.com/cokelaer/bioservices/issues',
    'description':'Access to Biological Web Services from Python' ,
    "long_description_content_type": "text/x-rst",
    'platforms' : ['Linux', 'Unix', 'MacOsX', 'Windows'],
    'keywords' : [
        "BioServices", "WebServices", "Biology", "BioDBNet",
        "ChEBI", "UniChem", "Kegg", "KEGG", "BioModels",
        "EUtils", "UniProt", "PICR", "ArrayExpress", "MUSCLE",
        "QuickGO", "PDB", "PSICQUIC", "Blast", "BioMART", "PantherDB",
        "BioGRID", "MIRIAM", "BioMart", "GeneProf", "ChEMBL",
        "ChemSpider",  "HGNC", "PathwayCommons", "Rhea", "Ensembl"],
    'classifiers' : [
          'Development Status :: 5 - Production/Stable',
          'Intended Audience :: Developers',
          'Intended Audience :: Science/Research',
          'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
          'Operating System :: OS Independent',
          'Programming Language :: Python :: 3.7',
          'Programming Language :: Python :: 3.8',
          'Programming Language :: Python :: 3.9',
          'Topic :: Software Development :: Libraries :: Python Modules',
          'Topic :: Scientific/Engineering :: Bio-Informatics',
          'Topic :: Scientific/Engineering :: Information Analysis',
          'Topic :: Scientific/Engineering :: Mathematics',
          'Topic :: Scientific/Engineering :: Physics']
    }


with open('README.rst', 'r', encoding='utf-8') as f:
    long_description =  f.read()

with open('requirements.txt', 'r') as f:
    install_requires = f.read()


on_rtd = os.environ.get('READTHEDOCS', None) == 'True'

# sphinx-gallery and numpydoc are used for the doc only.
# Could have a if on_rtd
#install_requires = ["grequests", "requests",
#        "requests_cache", "easydev>=0.9.36", "beautifulsoup4", "xmltodict",
#        "lxml",
#        "suds-community", "appdirs", 'wrapt', "pandas", "colorlog"],

setup(
    name             = 'bioservices',
    version          = version,
    maintainer       = metainfo['authors']['Cokelaer'][0],
    maintainer_email = metainfo['authors']['Cokelaer'][1],
    author           = metainfo['authors']['Cokelaer'][0],
    author_email     = metainfo['authors']['Cokelaer'][1],
    long_description = long_description, 
    long_description_content_type = metainfo["long_description_content_type"],
    keywords         = metainfo['keywords'],
    description = metainfo['description'],
    license          = metainfo['license'],
    platforms        = metainfo['platforms'],
    url              = metainfo['url'],
    download_url     = metainfo['download_url'],
    classifiers      = metainfo['classifiers'],

    # package installation
    package_dir = {'':'src'},
    packages = ['bioservices', 'bioservices.apps', 'bioservices.mapping'],
    #package_dir  = package_dir,

    # If user of python2.6 ordereddict must be installed manually
    install_requires = install_requires,
    extras_require={
        "testing": [
            "pytest",
            "pytest-cov",
            "pytest-xdist",
            "pytest-mock",
            "pytest-timeout",
            "pytest-runner",
            "coveralls",
        ],
    },
    entry_points = {
        'console_scripts':[
           'bioservices=bioservices.main:main',
        ]
    }

    )





#
#  This file is part of bioservices software
#
#  Copyright (c) 2013-2014 - EBI-EMBL
#
#  File author(s):
#      Thomas Cokelaer <cokelaer@ebi.ac.uk>
#
#  Distributed under the GPLv3 License.
#  See accompanying file LICENSE.txt or copy at
#      http://www.gnu.org/licenses/gpl-3.0.html
#
#  website: https://github.com/cokelaer/bioservices
#  documentation: http://packages.python.org/bioservices
#
##############################################################################
"""Interface to the WikiPathway service

.. topic:: What is WikiPathway ?

    :URL: http://www.wikipathways.org/index.php/WikiPathways
    :REST: http://webservice.wikipathways.org/
    :Citation: `doi:10.1371/journal.pone.0006447 <http://www.plosone.org/article/info:doi/10.1371/journal.pone.0006447>`_

    .. highlights::

        " WikiPathways is an open, public platform dedicated to the curation of
        biological pathways by and for the scientific community."

        -- From WikiPathway web site. Dec 2012

"""
from bioservices.services import REST
import copy, webbrowser, base64

import pandas as pd

__all__ = ["WikiPathways"]


class WikiPathways(REST):
    """Interface to `Pathway <http://www.wikipathways.org/index.php>`_ service

    .. doctest::

       >>> from bioservices import WikiPathways
       >>> s = Wikipathway()
       >>> s.organism  # default organism
       'Homo sapiens'

    Examples::

        s.findPathwaysByText('MTOR')
        s.getPathway('WP1471')
        s.getPathwaysByOntologyTerm('DOID:344')
        s.findPathwaysByXref('P45985')

    The methods that require a login are not implemented (:meth:`login`,
    :meth:`updatePathway`, :meth:`removeCurationTag`, :meth:`saveCurationTag`,
    :meth:`createPathway`)

    Methods not implemented at all:

      * u'getCurationTagHistory': No API found in Wikipathway web page
      * u'getRelations': No API found in Wikipathway web page
    """

    _url = "http://webservice.wikipathways.org/"

    def __init__(self, verbose=True, cache=False):
        """.. rubric:: Constructor

        :param bool verbose:

        """
        super(WikiPathways, self).__init__(name="WikiPathways", url=WikiPathways._url, verbose=verbose, cache=cache)
        self._organism = "Homo sapiens"  # This function is redundant (see class service)
        self.logging.info("Fetching organisms...")

        #: Get a list of all available organisms.
        self.organisms = self.listOrganisms()

    def listOrganisms(self):
        res = self.http_get(self.url + "/listOrganisms?format=json")
        return res["organisms"]

    def _set_organism(self, organism):
        if organism in self.organisms:
            self._organism = organism
        else:
            raise ValueError("'%s' is not supported in WikiPathways. See :attr:`organisms`" % organism)

    def _get_organism(self):
        return self._organism

    organism = property(_get_organism, _set_organism, doc="Read/write attribute for the organism")

    def findPathwaysByLiterature(self, query):
        """Find pathways by their literature references.

        :param str query: The query, can be a pubmed id, author name
            or title keyword.
        :return:  dictionary with Pathway as keys

        ::

            res = s.findPathwaysByLiterature(18651794)

        """
        params = {"format": "json", "query": query}
        res = self.http_get(self.url + "findPathwaysByLiterature", params=params)

        return res["result"]

    def findPathwaysByXref(self, ids, codes=None):
        """Find pathways by searching on the external references of DataNodes.

        :param str string ids: One or mode DataNode identifier(s) (e.g. 'P45985').
            Datanodes can be (gene/protein/metabolite identifiers). For one
            node, you can use a string (or number) or list of one identifier.
            you can also provide a list of identifiers.
        :param str codes: You can restrict the search to a specific database.
            See http://developers.pathvisio.org/wiki/DatabasesMapps#Supporteddatabasesystems
            for details. Examples are "L" for entrez gene, "En" for ensembl. See
            also the note here below for multiple identifiers/codes.
        :return:  a dictionary

        ::

            >>> s.findPathwaysByXref(ids="P45985")
            >>> s.findPathwaysByXref(ids="P45985", codes="L")
            >>> s.findPathwaysByXref(ids=["P45985"], codes=["L"])
            >>> s.findPathwaysByXref(ids=["P45985", "ENSG00000130164"], codes=["L", "En"])


        Note that in the last example, we specify multiple ids and codes
        parameters to query for multiple xrefs at once. In that case, the
        number of ids and codes parameters should match. Moreover, they will
        be paired to form xrefs, so P45985 is searched for in the "L"
        database while "ENSG00000130164" is searched for in the En" database
        only.
        """
        url = self.url + "/findPathwaysByXref?ids="
        if isinstance(ids, (str, int, float)):
            url += "{}".format(ids)
        elif isinstance(ids, list):
            if len(ids) == 0:
                raise ValueError("ids must be a non-empty list")
            if len(ids) >= 1:
                url += "{}".format(ids[0])
            if len(ids) > 1:
                for this in ids[1:]:
                    url += "&ids={}".format(this)
        else:
            raise ValueError("ids must be a list, or a string or a number")

        if codes and isinstance(codes, (str, int)):
            codes = [codes]
        if codes:
            for code in codes:
                url += "&codes={}".format(code)
        res = self.http_get(url + "&format=json")

        # results = pd.DataFrame(results)

        return res

    def findInteractions(self, query):
        """Find interactions defined in WikiPathways pathways.


        :param str query:  The name of an entity to find interactions for (e.g. 'P53')
        :returns: list of dictionaries

        ::

            res = w.findInteractions("P53")

        """
        url = self.url + "findInteractions?query={}&format=json".format(query)
        res = self.http_get(url)["result"]

    def listPathways(self, organism=None):
        """Get a list of all available pathways.

        :param str organism: If provided, the data is filtered to keep only
            the organism  provided, which must be a valid name (check out
            :attr:`organism` attribute)
        :return: dataframe. Index are the pathways identifiers (e.g. WP1)

        .. plot::

            from bioservices import WikiPathways
            w = WikiPathways()
            df = w.listPathways()
            df.groupby("species").count()['name'].sort_values().plot(kind="barh")
        """
        if organism:
            self.devtools.check_param_in_list(organism, self.organisms)
            request = self.http_get("/listPathways?%s&format=json" % organism)
        else:
            request = self.http_get("/listPathways?format=json")
        pathways = request["pathways"]

        pathways = pd.DataFrame(pathways).set_index("id")
        return pathways

    def getPathway(self, pathwayId, revision=0):
        """Download a pathway from WikiPathways.

        :param str pathwayId: the pathway identifier.
        :param int revision: the revision number of the pathway (use '0'
            for most recent version).
        :Returns: The pathway as a dictionary. The pathway is stored in gpml
            format.

        ::

            s.getPathway("WP2320")
        """
        url = "getPathway?pwId=%s" % pathwayId
        url += "&revision=%s&format=json" % revision
        request = self.http_get(url)

        return request["pathway"]

    def getPathwayInfo(self, pathwayId):
        """Get some general info about the pathway.

        :param str pathwayId: the pathway identifier.
        :return: The pathway info.

        ::

            >>> from bioservices import *
            >>> s = Wikipathway()
            >>> s.getPathwayInfo("WP2320")
        """
        url = "/getPathwayInfo?pwId=%s" % pathwayId
        request = self.http_get(url)
        data = self.easyXML(request.content)
        data = data.findAll("ns1:pathwayinfo")
        pathway = {}
        for this in data:
            for tag in ["id", "url", "name", "species", "revision"]:
                text = this.find("ns2:%s" % tag).getText()
                pathway[tag] = text
        return pathway

    def getPathwayHistory(self, pathwayId, date):
        """Get the revision history of a pathway.

        :param str pathwayId: the pathway identifier.
        :param str date: limit the results by date, only history items after
            the given date (timestamp format) will be included. Can be a string
            or number of the form YYYYMMDDHHMMSS.
        :return: The revision history.

        .. warning:: seems unstable does not return the results systematically.

        ::

            s.getPathwayHistory("WP4", 20110101000000)

        """

        query = self.url + "/getPathwayHistory?pwId=%s&timestamp=%s" % (
            pathwayId,
            str(date),
        )
        query += "&format=json"
        return self.http_get(query)

    def getRecentChanges(self, timestamp):
        """Get the recently changed pathways.

        :param str timestamp: Only get changes from after this time. Timestamp
            format: yyyymmddMMHHSS (string or number)
        :return: The changed pathways in XML format

        ::

            s.getRecentChanges(20110101000000)

        .. todo:: interpret XML
        """
        res = self.http_get(self.url + "/getRecentChanges?timestamp=%s&format=json" % timestamp)
        return res

    def login(self, usrname, password):
        """Start a logged in session using an existing WikiPathways account.

        .. warning:: Interface not exposed in bioservices.

        This function will return an authentication code that can
        be used to excecute methods that need authentication (e.g.
        updatePathway).

        :param str name: The username of the WikiPathways account.
        :param str password: The password of the WikiPathways account.


        :Returns: The authentication code for this session.

        """
        raise NotImplementedError
        # for future usage. pass is a python keyword so we must use a dictionary
        # d = {"name":usrname, "pass":password}
        # return self.serv.login(**d)

    def getPathwayAs(self, pathwayId, filetype="png", revision=0):
        """Download a pathway in the specified file format.

        :param str pathwayId: the pathway identifier.
        :param str filetype: the file format (default is .owl).
        :param int revision: the revision number of the pathway (use '0' for most recent version - this is default).
        :return: The file contents

        .. versionchanged:: 1.3.0 return raw output of the service without any parsing

        .. note:: use :meth:`savePathwayAs` to save into a file.
        """
        self.devtools.check_param_in_list(filetype, ["gpml", "png", "svg", "pdf", "txt", "pwf", "owl"])

        url = self.url + "/getPathwayAs?fileType=%s" % filetype
        url += "&pwId=%s " % pathwayId
        url += "&revision=%s&format=json" % revision
        res = self.http_get(url)
        return res["data"]

    def savePathwayAs(self, pathwayId, filename, revision=0, display=True):
        """Save a pathway.

        :param str pathwayId: the pathway identifier.
        :param str filename: the name of the file. If a filename extension
            is not provided the pathway will be saved as a pdf (default).
        :param int revisionNumb: the revision number of the pathway (use
            '0 for most recent version).
        :param bool display: if True the pathway will be displayed in your
            browser.

        .. note:: Method from bioservices. Not a WikiPathways function
        .. versionchanged:: 1.7 return PNG by default instead of PDF. PDF
            not working as of 20 Feb 2020 even on wikipathway website.
        """
        if filename.find(".") == -1:
            filename = "%s.%s" % (filename, "pdf")
        filetype = filename.split(".")[-1]

        res = self.getPathwayAs(pathwayId, filetype=filetype, revision=revision)

        with open(filename, "wb") as f:
            import binascii

            try:
                # python3
                newres = binascii.a2b_base64(bytes(res, "utf-8"))
            except:
                newres = binascii.a2b_base64(res)
            f.write(newres)

        if display:
            webbrowser.open(filename)
        f.close()

    def updatePathway(self, pathwayId, describeChanges, gpmlCode, revision=0):
        """Update a pathway on WikiPathways website with a given GPML code.

        .. warning:: Interface not exposed in bioservices.

        .. note:: To create/modify pathways via the web service, you need to
            have an account with web service write permissions. Please contact
            us to request write access for the web service.

        :param str pwId:  The pathway identifier.
        :param str description:  A description of the modifications.
        :param str gpml: The updated GPML code.
        :param int revision: The revision number of the version this GPML
            code was based on. This is used to prevent edit conflicts in
            case another client edited the pathway after this client downloaded it.
        :param object WSAuth_auth:  The authentication info.

        :returns: Boolean. True if the pathway was updated successfully.
        """
        raise NotImplementedError
        # return self.serv.updatePathway(pwId = pathwayId,
        # description = describeChanges, gpml = gpmlCode, revision = revisionNumb, auth = authInfo)

    def createPathway(self, gpmlCode, authInfo):
        """Create a new pathway on the WikiPathways website with a given GPML code.

        .. warning:: Interface not exposed in bioservices.

        .. note:: To create/modify pathways via the web service, you need to
            have an account with web service write permissions. Please
            contact us to request write access for the web service.

        :param str gpml: The GPML code.
        :param object WSAuth auth: The authentication info.
        :returns: WSPathwayInfo The pathway info for the created pathway
            (containing identifier, revision, etc.).

        """
        raise NotImplementedError
        # return self.serv.createPathway(gpml = gpmlCode, auth = authInfo)

    def saveCurationTag(self, pathwayId, name, revision):
        """Apply a curation tag to a pathway. This operation will overwrite any existing tag with the same name.

        .. warning:: Interface not exposed in bioservices.

        :param str pathwayId: the pathway identifier.
        """
        raise NotImplementedError

    def removeCurationTag(self, pathwayId, name):
        """Remove a curation tag from a pathway.

        .. warning:: Interface not exposed in bioservices.

        """
        raise NotImplementedError

    def getCurationTags(self, pathwayId):
        """Get all curation tags for the given pathway.

        :param str pathwayId: the pathway identifier.
        :returns: Array of WSCurationTag. The curation tags.

        ::

            s.getCurationTags("WP4")
        """
        raise NotImplementedError

    def getCurationTagsByName(self, name):
        """Get all curation tags for the given tag name.

        Use this method if you want to find all pathways that are tagged with a specific curation tag.


        :param str tagName: The tag name.
        :return: Array of WSCurationTag. The curation tags (one instance for each pathway that has been tagged).

        ::

            s.getCurationTagsByName("Curation:FeaturedPathway")
        """
        raise NotImplementedError

    def getColoredPathway(self, pathwayId, filetype="svg", revision=0, color=None, graphId=None):
        """Get a colored image version of the pathway.

        :param str pwId: The pathway identifier.
        :param int revision: The revision number of the pathway (use '0' for most recent version).
        :param str fileType:  The image type (One of 'svg', 'pdf' or 'png'). Not
            yet implemented. svg is returned for now.
        :returns: Binary form of the image.

        .. todo:: graphId, color parameters
        """

        url = self.url + "getColoredPathway?pwId={}".format(pathwayId)
        if revision:
            url += "&revision={}".format(revision)
        url += "&format=json"
        request = self.http_get(url)
        try:
            data = request["data"]
            return base64.b64decode(data)
        except:
            return request

    def findPathwaysByText(self, query, species=None):
        """Find pathways using a textual search on the description and text labels of the pathway objects.

        The query syntax offers several options:

        * Combine terms with AND and OR. Combining terms with a space is equal
          to using OR ('p53 OR apoptosis' gives the same result as 'p53 apoptosis').
        * Group terms with parentheses, e.g. '(apoptosis OR mapk) AND p53'
        * You can use wildcards * and ?. * searches for one or more
          characters, ? searches for only one character.
        * Use quotes to escape special characters. E.g. '"apoptosis*"' will
          include the * in the search and not use it as wildcard.

        This function supports REST-style invocation.
        Example: http://www.wikipathways.org/wpi/webservice/webservice.php/findPathwaysByText?query=apoptosis

        :param str query: The search query (e.g. 'apoptosis' or 'p53').
        :param str species:  The species to limit the search to (leave blank to search on all species).
        :return: Array of WSSearchResult An array of search results.

        ::

            s.findPathwaysByText(query="p53 OR mapk",species='Homo sapiens')

        .. warning:: AND or OR must be in big caps
        """
        url = self.url + "/findPathwaysByText?query=%s" % query
        if species:
            url += "&species=%s" % species
        url += "&format=json"
        request = self.http_get(url)
        data = request["result"]

        try:
            data = pd.DataFrame(data).set_index("id")
        except:
            pass
        return data

    def getOntologyTermsByPathway(self, pathwayId):
        """Get a list of ontology terms for a given pathway.


        :param str pathwayId: the pathway identifier.
        :return: Array of WSOntologyTerm. The ontology terms.

        ::

            s.getOntologyTermsByPathway("WP4")
        """
        url = self.url + "getOntologyTermsByPathway?pwId={}".format(pathwayId)
        url += "&format=json"
        request = self.http_get(url)
        results = request["terms"]

        return results

    def getPathwaysByOntologyTerm(self, terms):
        """Get a list of pathways tagged with a given ontology term.

        :param str terms: the ontology term identifier.
        :returns: dataframe with pathways infomation.

        ::

            >>> from bioservices import WikiPathways
            >>> s = Wikipathway()
            >>> s.getPathwaysByOntologyTerm('PW:0000724')

        """
        url = self.url + "getPathwaysByOntologyTerm?term={}".format(terms)
        url += "&format=json"
        request = self.http_get(url)
        return request["pathways"]

    def getPathwaysByParentOntologyTerm(self, term):
        """Get a list of pathways tagged with any ontology term that is the child of the given Ontology term.

        :param str term: the ontology term identifier.
        :returns: List of WSPathwayInfo The pathway information.

        """
        url = self.url + "getPathwaysByParentOntologyTerm?term={}".format(term)
        url += "&format=json"
        request = self.http_get(url)
        return request["pathways"]

    def showPathwayInBrowser(self, pathwayId):
        """Show a given Pathway into your favorite browser.

        :param str pathwayId: the pathway identifier.

        .. note: This is an additional bioservices functionality (not a
            wikipathway one) showing a wikipathway URL.

        """
        url = self.getPathwayInfo(pathwayId)["url"]
        webbrowser.open(url)



#  This file is part of bioservices software
#
#  Copyright (c) 2013-2014 - EBI-EMBL
#  Copyright (c) 2021 - Institut Pasteur
#
#  File author(s):
#      Thomas Cokelaer <cokelaer@ebi.ac.uk>
#
#  Distributed under the GPLv3 License.
#  See accompanying file LICENSE.txt or copy at
#      http://www.gnu.org/licenses/gpl-3.0.html
#
#  website: https://github.com/cokelaer/bioservices
#  documentation: http://packages.python.org/bioservices
#
##############################################################################
"""Interface to biocontainer

.. topic:: What is biocontainers

    :URL: https://biocontainers.pro/
    :Citation:

    .. highlights::

        BioContainers is an open-source project that aims to create,
        store, and distribute bioinformatics software containers and
        packages.

        -- From biocontainers (about), Jan 2021

"""
import pandas as pd

from bioservices.services import REST
from bioservices import logger

logger.name = __name__


__all__ = ["Biocontainers"]


class Biocontainers:
    """Interface to Biocontainers service

    ::

        >>> from bioservics import Biocontainers
        >>> b = Biocontainers()
        >>> b.get_tools()

    """

    _url = "https://api.biocontainers.pro/ga4gh/trs/v2/"

    def __init__(self, verbose=True, cache=False):
        """**Constructor**

        :param verbose: set to False to prevent informative messages
        """
        self.services = REST(
            name="biocontainers",
            url=Biocontainers._url,
            verbose=verbose,
            cache=cache,
            url_defined_later=True,
        )

    def get_tools(self, limit=20000):
        """Returns all available tools."""
        params = {"limit": limit, "sort_ield": "id", "sort_order": "asc"}
        res = self.services.http_get("tools", params=params)
        try:
            return pd.DataFrame(res)
        except Exception:
            return res

    def get_stats(self):
        """Returns some stats about numer of versions and tools"""
        res = self.services.http_get("stats")
        return res

    def get_versions_one_tool(self, tool):
        """Returns all versions of a given tool."""
        res = self.services.http_get(f"tools/{tool}/versions", params={})
        try:
            return pd.DataFrame(res)
        except Exception:
            return res
